#ifndef CEGAR_SPLIT_SELECTOR_H
#define CEGAR_SPLIT_SELECTOR_H

#include "../task_proxy.h"

#include <memory>
#include <vector>

namespace additive_heuristic {
class AdditiveHeuristic;
}

namespace utils {
class RandomNumberGenerator;
}

namespace cegar {
class Abstraction;
class AbstractState;

// Strategies for selecting a split in case there are multiple possibilities.
enum class PickSplit {
    RANDOM,
    // Number of values that land in the state whose h-value is probably raised.
    MIN_UNWANTED,
    MAX_UNWANTED,
    // Refinement: - (remaining_values / original_domain_size)
    MIN_REFINED,
    MAX_REFINED,
    // Compare the h^add(s_0) values of the facts.
    MIN_HADD,
    MAX_HADD,
    // Compare the causal graph index.
    MIN_CG,
    MAX_CG,
    // Compare the average goal distance resulting from the split.
    MIN_GOAL_DIST,
    MAX_GOAL_DIST,
    // Compare how many abstract states have a higher goal distance
    MIN_HIGHER_DIST,
    MAX_HIGHER_DIST,
    // Count the state-change-inducing operators of the abstraction
    MIN_ACTIVE_OPS,
    MAX_ACTIVE_OPS
};


struct Split {
    int var_id;
    std::vector<int> values;

    Split(int var_id, std::vector<int> &&values)
        : var_id(var_id), values(move(values)) {
    }
};


/*
  Select split in case there are multiple possible splits.
*/
class SplitSelector {
    const std::shared_ptr<AbstractTask> task;
    const TaskProxy task_proxy;
    std::unique_ptr<additive_heuristic::AdditiveHeuristic> additive_heuristic;
    std::unique_ptr<std::vector<int>> op_costs;

    const PickSplit pick;
    unsigned int num_picks;
    unsigned long long num_options;
    unsigned long long num_ratings;

    int get_num_unwanted_values(const AbstractState &state, const Split &split) const;
    double get_refinedness(const AbstractState &state, int var_id) const;
    int get_hadd_value(int var_id, int value) const;
    int get_min_hadd_value(int var_id, const std::vector<int> &values) const;
    int get_max_hadd_value(int var_id, const std::vector<int> &values) const;
    double get_avg_goal_dist(const Abstraction &abstraction, const AbstractState &state, const Split &split) const;
    int get_num_inc_goal_dist(const Abstraction &abstraction, const AbstractState &state, const Split &split) const;
    size_t get_num_active_ops(const Abstraction &abstraction, const AbstractState &state, const Split &split) const;

    double rate_split(const Abstraction &abstraction, const AbstractState &state, const Split &split) const;

public:
    SplitSelector(const std::shared_ptr<AbstractTask> &task, PickSplit pick);
    ~SplitSelector();

    const Split &pick_split(
        const Abstraction &abstraction,
        const AbstractState &state,
        const std::vector<Split> &splits,
        utils::RandomNumberGenerator &rng);

    void print_statistics() const;
};
}

#endif
